// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SOUNDWIN.CC) Copyright 1998-1999 by Josef Drexler
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com
//
// For info on this file please see sound.html in the docs directory
//
// This file implements the sound output to the Windows Midi Mapper for music
// and the Wave Mapper for audio output.
//
// There are two way to get audio output, one through the wave mapper and the
// other by using sndPlaySnd. The latter is not actually working at the moment
// but can be activated by undefining usewaveOut and defining usesndPlaySnd instead
//


#include "bochs.h"
#include <windows.h>


#if BX_USE_SB16_SMF
#define this NULL
#endif

// uncomment one of the following two #defines
//#define usesndPlaySnd
#define usewaveOut

// Definitions for WINMM.DLL, if not defined already
#ifndef MMSYSERR_NOERROR

typedef UINT HMIDIOUT;
typedef	HMIDIOUT *LPHMIDIOUT;
typedef struct midihdr_tag {
    LPSTR lpData;
    DWORD dwBufferLength;
    DWORD dwBytesRecorded;
    DWORD dwUser;
    DWORD dwFlags;
    struct midihdr_tag *lpNext;
    DWORD reserved;
} MIDIHDR, *LPMIDIHDR;

typedef UINT HWAVEOUT;
typedef HWAVEOUT *LPHWAVEOUT;

typedef struct wavehdr_tag {
    LPSTR lpData;
    DWORD dwBufferLength;
    DWORD dwBytesRecorded;
    DWORD dwUser;
    DWORD dwFlags;
    DWORD dwLoops;
    struct wavehdr_tag *lpNext;
    DWORD reserved;
} WAVEHDR, *LPWAVEHDR;

#define WHDR_DONE         0x00000001
#define WHDR_PREPARED     0x00000002
#define WHDR_BEGINLOOP    0x00000004
#define WHDR_ENDLOOP      0x00000008
#define WHDR_INQUEUE      0x00000010

typedef struct waveformat_tag {
    WORD wFormatTag;
    WORD nChannels;
    DWORD nSamplesPerSec;
    DWORD nAvgBytesPerSec;
    WORD nBlockAlign;
    WORD nExtraData;
} WAVEFORMAT, *LPWAVEFORMAT;

#define WAVE_FORMAT_PCM 	1

typedef struct pcmwaveformat_tag {
    WAVEFORMAT wf;
} PCMWAVEFORMAT, *LPPCMWAVEFORMAT;

#define MIDIMAPPER		-1
#define WAVEMAPPER		-1

#define CALLBACK_NULL		0x00000000
#define CALLBACK_WINDOW		0x00010000
#define CALLBACK_TASK		0x00020000
#define CALLBACK_FUNCTION	0x00030000

#define MMSYSERR_NOERROR	0
#define MMSYSERR_ERROR		1
#define MMSYSERR_BADDEVICEID	2
#define MMSYSERR_NOTENABLED	3
#define MMSYSERR_ALLOCATED      4
#define MMSYSERR_INVALHANDLE	5
#define MMSYSERR_NODRIVER	6
#define MMSYSERR_NOMEM		7
#define MMSYSERR_NOTSUPPORTED	8
#define MMSYSERR_NOMAP		7

#define MIDIERR_UNPREPARED      64
#define MIDIERR_STILLPLAYING    65
#define MIDIERR_NOTREADY	66
#define MIDIERR_NODEVICE        67

#define WAVERR_BADFORMAT        32
#define WAVERR_STILLPLAYING     33
#define WAVERR_UNPREPARED       34
#define WAVERR_SYNC             35

#define MAXERRORLENGTH		128

extern "C" {
UINT STDCALL midiOutOpen(LPHMIDIOUT, UINT, DWORD, DWORD, DWORD);
UINT STDCALL midiOutShortMsg(HMIDIOUT, DWORD);
UINT STDCALL midiOutLongMsg(HMIDIOUT, LPMIDIHDR, UINT);
UINT STDCALL midiOutPrepareHeader(HMIDIOUT, LPMIDIHDR, UINT);
UINT STDCALL midiOutUnprepareHeader(HMIDIOUT, LPMIDIHDR, UINT);
UINT STDCALL midiOutReset(HMIDIOUT);
UINT STDCALL midiOutClose(HMIDIOUT);

UINT STDCALL waveOutOpen(LPHWAVEOUT, UINT, LPWAVEFORMAT, DWORD, DWORD, DWORD);
UINT STDCALL waveOutWrite(HWAVEOUT, LPWAVEHDR, UINT);
UINT STDCALL waveOutPrepareHeader(HWAVEOUT, LPWAVEHDR, UINT);
UINT STDCALL waveOutUnprepareHeader(HWAVEOUT, LPWAVEHDR, UINT);
UINT STDCALL waveOutReset(HWAVEOUT);
UINT STDCALL waveOutClose(HWAVEOUT);

UINT STDCALL waveOutGetErrorTextA(UINT, LPSTR, UINT);

BOOL STDCALL sndPlaySoundA(LPCSTR, UINT);
}

typedef struct {
  char RIFF[4];
  Bit32u length;
  char TYPE[4];
  char chnk[4];
  Bit32u chnklen;
  PCMWAVEFORMAT waveformat;
  char chnk2[4];
  Bit32u chnk2len;
  char data[1];
} WAVEFILEHEADER, *LPWAVEFILEHEADER;

#endif  // MMSYSERR_NOERROR defined


#define BX_SB16_OUTNBUF 	4   // number of buffers for the output, must be power of 2
#define BX_SB16_OUTNMASK	(BX_SB16_OUTNBUF - 1)

struct bx_sb16_waveinfo_struct {
  int frequency;
  int bits;
  int stereo;
  int format;
};

struct bx_sb16_output_struct {
  HMIDIOUT MidiOut;       // Midi output device
  int MidiOpen;           // is it open?

  HWAVEOUT WaveOut;       // Wave output device
  int WaveOpen;           // is it open?

  UINT WaveDevice;        // Wave device ID, for waveOutOpen

       // some data for the buffers
  HANDLE WaveHeaderHandle[BX_SB16_OUTNBUF];     // returned by GlobalAlloc()
  LPWAVEHDR WaveHeader[BX_SB16_OUTNBUF];        // returned by GlobalLock
  HANDLE WaveDataHandle[BX_SB16_OUTNBUF];
  LPSTR WaveData[BX_SB16_OUTNBUF];
  int length[BX_SB16_OUTNBUF];                  // length of the data in the buffer
  int needreopen;                               // if the format has changed
  int head,tailfull,tailplay;       // These are for three states of the buffers: empty, full, played
  bx_sb16_waveinfo_struct WaveInfo;             // format for the next buffer to be played
};

#define OUTPUTDATA   ( (bx_sb16_output_struct*) (BX_SB16_THIS outputdata) )


int bx_sb16_c::initoutput()
{
  if (OUTPUTDATA == NULL)
    OUTPUTDATA = new bx_sb16_output_struct;

  if (OUTPUTDATA == NULL)
    {
	// allocation of struct failed, disable sound output
      bx_sb16.writelog( BOTHLOG(2), "Couldn't allocate outputdata in soundwin.cc. Sound disabled.");
      bx_options.sb16.midimode = 0;
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  OUTPUTDATA->MidiOpen = 0;
  OUTPUTDATA->WaveOpen = 0;
  OUTPUTDATA->WaveDevice = WAVEMAPPER;

  // initially the output is ready
  BX_SB16_THIS waveoutputready = 1;
  BX_SB16_THIS midioutputready = 1;

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::doneoutput()
{
  if (OUTPUTDATA != NULL)
    delete(OUTPUTDATA);
  OUTPUTDATA = NULL;

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::openmidioutput(char *device)
{
  // could make the output device selectable,
  // but currently only the midi mapper is supported
  UNUSED(device);

  UINT deviceid = MIDIMAPPER;

  OUTPUTDATA->MidiOpen = 0;

  UINT ret = midiOutOpen( &(OUTPUTDATA->MidiOut), deviceid, NULL, NULL, CALLBACK_NULL);
  if (ret == 0)
    OUTPUTDATA->MidiOpen = 1;

  bx_sb16.writelog( MIDILOG(4), "midiOutOpen() = %d, MidiOpen: %d", ret, OUTPUTDATA->MidiOpen);

  return (OUTPUTDATA->MidiOpen == 1) ? BX_SB16_OUTPUT_OK : BX_SB16_OUTPUT_ERR;
}

int bx_sb16_c::sendmidicommand(int delta, int command, int length, Bit8u data[])
{
  if (OUTPUTDATA->MidiOpen != 1)
    return BX_SB16_OUTPUT_ERR;

  if ( (command == 0xf0) || (command == 0xf7) )
    {
      // need to do:
      // prepareheader etc.
      // midioutlongmsg...
      // midioutputready = 0 until command has been sent
      bx_sb16.writelog( MIDILOG(2), "sendmidicommand: SYSEX messages not supported (yet)");
      return BX_SB16_OUTPUT_ERR;
    }

  if (length > 3)
    {
      bx_sb16.writelog( MIDILOG(2), "sendmidicommand: message %d too long (%d)", command, length);
      return BX_SB16_OUTPUT_ERR;
    }
  DWORD msg = command;

  for (int i = 0; i<length; i++)
    msg |= (data[i] << (8 * (i + 1) ) );

  UINT ret = midiOutShortMsg(OUTPUTDATA->MidiOut, msg);
  bx_sb16.writelog( MIDILOG(4), "midiOutShortMsg(%x) = %d", msg, ret);

  return (ret == 0) ? BX_SB16_OUTPUT_OK : BX_SB16_OUTPUT_ERR;
}

int bx_sb16_c::closemidioutput()
{
  if (OUTPUTDATA->MidiOpen != 1)
    return BX_SB16_OUTPUT_ERR;

  UINT ret = midiOutClose(OUTPUTDATA->MidiOut);
  bx_sb16.writelog( MIDILOG(4), "midiOutClose() = %d", ret);
  OUTPUTDATA->MidiOpen = 0;

  return (ret == 0) ? BX_SB16_OUTPUT_OK : BX_SB16_OUTPUT_ERR;
}

int bx_sb16_c::openwaveoutput(char *device)
{
  // could make the output device selectable,
  // but currently only the midi mapper is supported
  UNUSED(device);

  bx_sb16.writelog( WAVELOG(4), "openwaveoutput(%s)", device);

#ifdef usewaveOut
  OUTPUTDATA->WaveDevice = WAVEMAPPER;

  for (int i=0; i<BX_SB16_OUTNBUF; i++)
    {
      OUTPUTDATA->WaveHeaderHandle[i] = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, sizeof(*OUTPUTDATA->WaveHeader[i]));
      OUTPUTDATA->WaveHeader[i] = (LPWAVEHDR) GlobalLock(OUTPUTDATA->WaveHeaderHandle[i]);

      if (OUTPUTDATA->WaveHeader[i] == NULL)
	{
	  bx_sb16.writelog( WAVELOG(2), "GlobalLock(Header) returned NULL pointer for handle %d", OUTPUTDATA->WaveHeaderHandle[i]);
	  bx_options.sb16.wavemode = 0;
	  return BX_SB16_OUTPUT_ERR;
	}

      OUTPUTDATA->WaveHeader[i]->dwFlags = WHDR_DONE;

      OUTPUTDATA->WaveDataHandle[i] = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, BX_SB16_WAVEPACKETSIZE);

      OUTPUTDATA->WaveData[i] = (LPSTR) GlobalLock(OUTPUTDATA->WaveDataHandle[i]);

      if (OUTPUTDATA->WaveData[i] == NULL)
	{
	  bx_sb16.writelog( WAVELOG(2), "GlobalLock(Data) returned NULL pointer");
	  bx_options.sb16.wavemode = 0;
	  return BX_SB16_OUTPUT_ERR;
	}
    }

  OUTPUTDATA->head = 0;
  OUTPUTDATA->tailfull = 0;
  OUTPUTDATA->tailplay = 0;
  OUTPUTDATA->needreopen = 0;

#endif

#ifdef usesndPlaySnd
  OUTPUTDATA->WaveDataHandle = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, BX_SB16_WAVEPACKETSIZE + 44);

  OUTPUTDATA->WaveData = GlobalLock(OUTPUTDATA->WaveDataHandle);

  if (OUTPUTDATA->WaveData == NULL)
    {
      bx_sb16.writelog( WAVELOG(2), "GlobalLock(Data) returned NULL pointer");
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }
#endif

  return BX_SB16_OUTPUT_OK;
}

int playnextbuffer(bx_sb16_output_struct *outputdata)
{
  UINT ret;
  PCMWAVEFORMAT waveformat;
  int bufnum;

  // if the format is different, we have to reopen the device,
  // so reset it first
  if (outputdata->needreopen != 0)
    if (outputdata->WaveOpen != 0)
      ret = waveOutReset( outputdata->WaveOut );

  // clean up all finished buffers and mark them available
  for (bufnum=outputdata->tailfull;
       (bufnum != outputdata->tailplay) &&
	  ( (outputdata->WaveHeader[bufnum]->dwFlags & WHDR_DONE) != 0);
       bufnum++, bufnum &= BX_SB16_OUTNMASK)
    ret = waveOutUnprepareHeader( outputdata->WaveOut, outputdata->WaveHeader[bufnum], sizeof(*outputdata->WaveHeader[bufnum]));
  outputdata->tailfull = bufnum;

  // enable gathering data if a buffer is available
  if (bufnum != outputdata->head)
    BX_SB16_THIS waveoutputready = 1;

  // do we have to play anything?
  if (outputdata->tailplay == outputdata->head)
    return BX_SB16_OUTPUT_OK;

  // if the format is different, we have to close and reopen the device
  // or, just open the device if it's not open yet
  if ( (outputdata->needreopen != 0) || (outputdata->WaveOpen == 0) )
    {
      if (outputdata->WaveOpen != 0)
	{
	  ret = waveOutClose( outputdata->WaveOut );
	  outputdata->WaveOpen = 0;
	}

      // try three times to find a suitable format
      for (int tries = 0; tries < 3; tries++)
	{
	  int frequency = outputdata->WaveInfo.frequency;
	  int stereo = outputdata->WaveInfo.stereo;
	  int bits = outputdata->WaveInfo.bits;
	  int format = outputdata->WaveInfo.format;
	  int bps = (bits / 8) * (stereo + 1);

	  waveformat.wf.wFormatTag = WAVE_FORMAT_PCM;
	  waveformat.wf.nChannels = stereo + 1;
	  waveformat.wf.nSamplesPerSec = frequency;
	  waveformat.wf.nAvgBytesPerSec = frequency * bps;
	  waveformat.wf.nBlockAlign = bps;
// stupid compiler can't align properly...
//	  waveformat.wBitsPerSample = bits;
	  waveformat.wf.nExtraData = bits;

	  ret = waveOutOpen( &(outputdata->WaveOut), outputdata->WaveDevice, &(waveformat.wf), NULL, NULL, CALLBACK_NULL);
	  if (ret != 0)
	    {
	      char errormsg[4*MAXERRORLENGTH+1];
	      waveOutGetErrorTextA(ret, errormsg, 4*MAXERRORLENGTH+1);
	      bx_sb16.writelog( WAVELOG(5), "waveOutOpen: %s", errormsg);
	      switch (tries)
		{
		case 0:	// maybe try a different frequency
		  if (frequency < 15600)
		    frequency = 11025;
		  else if (frequency < 31200)
		    frequency = 22050;
		  else
		    frequency = 44100;

		  bx_sb16.writelog( WAVELOG(4), "Couldn't open wave device (error %d), trying frequency %d", ret, frequency);

		  break;
		case 1:	// or something else
		  frequency = 11025;
		  stereo = 0;
		  bits = 8;
		  bps = 1;

		  bx_sb16.writelog( WAVELOG(4), "Couldn't open wave device again (error %d), trying 11KHz, mono, 8bit", ret);

		  break;
		case 2:	// nope, doesn't work

		  bx_sb16.writelog( WAVELOG(2), "Couldn't open wave device (error %d)!", ret);

		  return BX_SB16_OUTPUT_ERR;
		}
	      bx_sb16.writelog( WAVELOG(5), "The format was: wFormatTag=%d, nChannels=%d, nSamplesPerSec=%d,",
			waveformat.wf.wFormatTag, waveformat.wf.nChannels, waveformat.wf.nSamplesPerSec);
	      bx_sb16.writelog( WAVELOG(5), "                nAvgBytesPerSec=%d, nBlockAlign=%d, wBitsPerSample=%d",
			waveformat.wf.nAvgBytesPerSec, waveformat.wf.nBlockAlign, waveformat.wf.nExtraData);

	    }
	  else
	    {
	      outputdata->WaveOpen = 1;
	      break;
	    }
	}
    }

  for (bufnum=outputdata->tailplay; bufnum != outputdata->head;
       bufnum++, bufnum &= BX_SB16_OUTNMASK, outputdata->tailplay=bufnum)
    {
      // prepare the wave header
      outputdata->WaveHeader[bufnum]->lpData = outputdata->WaveData[bufnum];
      outputdata->WaveHeader[bufnum]->dwBufferLength = outputdata->length[bufnum];
      outputdata->WaveHeader[bufnum]->dwBytesRecorded = outputdata->length[bufnum];
      outputdata->WaveHeader[bufnum]->dwUser = 0;
      outputdata->WaveHeader[bufnum]->dwFlags = 0;
      outputdata->WaveHeader[bufnum]->dwLoops = 1;

      ret = waveOutPrepareHeader(outputdata->WaveOut, outputdata->WaveHeader[bufnum], sizeof(*outputdata->WaveHeader[bufnum]));
      if (ret != 0)
	{
	  bx_sb16.writelog( WAVELOG(2), "waveOutPrepareHeader = %d", ret);
	  return BX_SB16_OUTPUT_ERR;
	}

      ret = waveOutWrite(outputdata->WaveOut, outputdata->WaveHeader[bufnum], sizeof(*outputdata->WaveHeader[bufnum]));
      if (ret != 0)
	{
	  char errormsg[4*MAXERRORLENGTH+1];
	  waveOutGetErrorTextA(ret, errormsg, 4*MAXERRORLENGTH+1);
	  bx_sb16.writelog( WAVELOG(5), "waveOutWrite: %s", errormsg);
	}
    }
}

int bx_sb16_c::startwaveplayback(int frequency, int bits, int stereo, int format)
{
  UINT ret;

  bx_sb16.writelog( WAVELOG(4), "startwaveplayback(%d, %d, %d, %x)", frequency, bits, stereo, format);

#ifdef usewaveOut
  // check if any of the properties have changed
  if ( (OUTPUTDATA->WaveInfo.frequency != frequency) ||
       (OUTPUTDATA->WaveInfo.bits != bits) ||
       (OUTPUTDATA->WaveInfo.stereo != stereo) ||
       (OUTPUTDATA->WaveInfo.format != format) )
    {
      OUTPUTDATA->needreopen = 1;

      // store the current settings to be used by sendwavepacket()
      OUTPUTDATA->WaveInfo.frequency = frequency;
      OUTPUTDATA->WaveInfo.bits = bits;
      OUTPUTDATA->WaveInfo.stereo = stereo;
      OUTPUTDATA->WaveInfo.format = format;
    }
#endif

#ifdef usesndPlaySnd
  int bps = (bits / 8) * (stereo + 1);
  LPWAVEFILEHEADER header = (LPWAVEFILEHEADER) OUTPUTDATA->WaveData;

  memcpy(header->RIFF, "RIFF", 4);
  memcpy(header->TYPE, "WAVE", 4);
  memcpy(header->chnk, "fmt ", 4);
  header->chnklen = 16;
  header->waveformat.wf.wFormatTag = WAVE_FORMAT_PCM;
  header->waveformat.wf.nChannels = stereo + 1;
  header->waveformat.wf.nSamplesPerSec = frequency;
  header->waveformat.wf.nAvgBytesPerSec = frequency * bps;
  header->waveformat.wf.nBlockAlign = bps;
//  header->waveformat.wBitsPerSample = bits;
  header->waveformat.wf.nExtraData = bits;
  memcpy(header->chnk2, "data", 4);
#endif

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::sendwavepacket(int length, Bit8u data[])
{
  UINT ret;
  int bufnum;

  bx_sb16.writelog( WAVELOG(4), "sendwavepacket(%d, %p)", length, data);

#ifdef usewaveOut

  bufnum = OUTPUTDATA->head;

  memcpy(OUTPUTDATA->WaveData[bufnum], data, length);
  OUTPUTDATA->length[bufnum] = length;

  // select next buffer to write to
  bufnum++;
  bufnum &= BX_SB16_OUTNMASK;

  if ( ( (bufnum + 1) & BX_SB16_OUTNMASK) == OUTPUTDATA->tailfull )
    {   // this should not actually happen!
      writelog( WAVELOG(2), "Output buffer overflow! Not played.");
      BX_SB16_THIS waveoutputready = 0;  // stop the output for a while
      return BX_SB16_OUTPUT_ERR;
    }

  OUTPUTDATA->head = bufnum;

  // check if more buffers are available, otherwise stall the emulator
  if ( ( (bufnum + 2) & BX_SB16_OUTNMASK) == OUTPUTDATA->tailfull )
    BX_SB16_THIS waveoutputready = 0;

  playnextbuffer(OUTPUTDATA);

#endif

#ifdef usesndPlaySnd
  LPWAVEFILEHEADER header = (LPWAVEFILEHEADER) OUTPUTDATA->WaveData;

  header->length = length + 36;
  header->chnk2len = length;

  memcpy( &(header->data), data, length);

  FILE *test = fopen("test", "a");
  fwrite(OUTPUTDATA->WaveData, 1, length + 44, test);
  fclose(test);

  ret = sndPlaySoundA( (LPCSTR) header, SND_SYNC | SND_MEMORY );
  if (ret != 0)
    {
      bx_sb16.writelog( WAVELOG(3), "sndPlaySoundA: %d", ret);
    }
#endif

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::stopwaveplayback()
{
  bx_sb16.writelog( WAVELOG(4), "stopwaveplayback()");

#ifdef usewaveOut
  // this is handled by the playnextbuffer when closing
#endif

#ifdef usesndPlaySnd
  sndPlaySoundA( NULL, SND_ASYNC | SND_MEMORY );

  OUTPUTDATA->WaveOpen = 0;
#endif

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::closewaveoutput()
{
  int bufnum;

  bx_sb16.writelog( WAVELOG(4), "closewaveoutput");

#ifdef usewaveOut
  if (OUTPUTDATA->WaveOpen == 1)
    {
      waveOutReset( OUTPUTDATA->WaveOut );

      // let playnextbuffer() clean up the buffers
      playnextbuffer(OUTPUTDATA);

      waveOutClose( OUTPUTDATA->WaveOut );

      OUTPUTDATA->head = 0;
      OUTPUTDATA->tailfull = 0;
      OUTPUTDATA->tailplay = 0;
      OUTPUTDATA->needreopen = 0;
    }

  for (bufnum=0; bufnum<BX_SB16_OUTNBUF; bufnum++)
    {
      GlobalUnlock(OUTPUTDATA->WaveHeaderHandle[bufnum]);
      GlobalFree(OUTPUTDATA->WaveHeaderHandle[bufnum]);

      GlobalUnlock(OUTPUTDATA->WaveDataHandle[bufnum]);
      GlobalFree(OUTPUTDATA->WaveDataHandle[bufnum]);
    }
#endif

#ifdef usesndPlaySnd
  GlobalUnlock(OUTPUTDATA->WaveDataHandle);
  GlobalFree(OUTPUTDATA->WaveDataHandle);
#endif

  return BX_SB16_OUTPUT_OK;
}
